/*
 * ccl-exception.h -- An exception mechanism for C.
 * 
 * This file is a part of the C Common Library (CCL) project. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the AltaRica Public License that comes with this
 * package.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 */

/*!
 * \file ccl/ccl-exception.h
 * \brief An exception mechanism.
 *
 * This module implements exceptions for ANSI C programs. The implementation is
 * based on the \c setjmp and \c longjmp standard functions. This code is 
 * inspired by the one of A. Braquelaire (http://www.labri.fr/~achille/).
 * 
 * \section definition Definition and declarations of exceptions
 *
 * A new exception kind, say \c ex, is created using the following statement 
 * generally at the top-level scope:
 * \code CCL_DEFINE_EXCEPTION(ex, parent); \endcode
 * \c ex is the name for the exception kind and \c parent is the name of
 * a parent kind. 
 * 
 * Exception are organized hierarchically. An exception of kind \c ex can be 
 * caught by any code that declares to catch \c ex exceptions or one of its 
 * ancestors in the hierarchy of exceptions.
 *
 * The macro creates a new symbol based on \c ex (generated by the macro \ref 
 * CCL_EXCEPTION_NAME) that should not be used directly. 
 *
 * The module pre-defines some kind of exceptions:
 * \li \ref ccl_exception_exception "exception"
 * \li \ref ccl_exception_error "error"
 * \li \ref ccl_exception_internal_error "internal_error" inherits from \ref 
 * ccl_exception_error "error"
 * \li \ref ccl_exception_runtime_exception "runtime_exception" inherits from 
 * \ref ccl_exception_exception "exception"
 * \li \ref ccl_exception_division_by_zero "division_by_zero" inherits from 
 * \ref ccl_exception_runtime_exception "runtime_exception"
 * 
 * Exceptions can be declared e.g. in header files using the following 
 * statement:
 * \code
 * CCL_DECLARE_EXCEPTION(ex, parent);
 * \endcode
 *
 * \section raise Raise of exceptions.
 * 
 * An exception can be raised (or thrown) for several reasons: detected errors, 
 * unexpected behaviors, algorithmic issues, ... 
 * Assume that there exists some kind of exception, say \e my_exception, then
 * the following code
 * \code
 * if (some_condition_is_true)
 *    ccl_throw(my_exception, "hum! something wrong occurs !");
 * \endcode
 * raised an exception of kind \e my_exception. The throw of the exception is 
 * accompagned with a explanation message that can be used by caller codes.
 * If no message has to be sent \ref ccl_throw_no_msg can be used in place of
 * \ref ccl_throw.
 * 
 * One can use the dummy macro-function \ref CCL_THROW to indicate after the 
 * declaration of a function that this latter raised exceptions. Example:
 * \code
 * extern int read_and_parse_line_of_file (const char *filename)
 *    CCL_THROW (io_exception  parser_exception);
 * \endcode
 * The \ref CCL_THROW generates no code.
 *
 * \section try Handling exceptions
 *
 * The \ref ccl_try(\e ex) statement is used to indicate that one wants to 
 * handle exceptions of kined \e ex raised by the next block of code. The block
 * of code handled by the \ref ccl_try statement is immediatly followed by:
 * \li a \ref ccl_catch statement if the exception has to be treated. This block
 *     has to be terminated with a \ref ccl_end_try statement.
 * \li a \ref ccl_no_catch statement if the exception is ignored.
 *
 * Example:
 * \code
 * extern int read_and_parse_line_of_file (const char *filename)
 *    CCL_THROW (io_exception  parser_exception);
 * 
 * int my_function(void)
 * {
 *    ccl_try (parser_exception)
 *      {
 *         read_and_parse_line_of_file ("file_to_read");
 *      }
 *    ccl_catch
 *      {
 *         ... code that treats the exception ...
 *      }
 *    ccl_end_try;
 * }
 * \endcode
 *
 * On this example, one can remark that the \ref ccl_try only handle exceptions
 * of kind \e parser_exception. That means that, if an \e io_exception occurs,
 * it is not handled by this code. If no treatment is done then the lines from
 * <tt>ccl_catch ... ccl_end_try</tt> can be replaced by <tt>ccl_no_catch;</tt>
 *
 * In a <tt>ccl_catch ... ccl_end_try</tt> block one can use \ref ccl_rethrow
 * to re-throw the last raised exception (this is useful to propagate error 
 * treatements).
 * 
 * \ref ccl_try can have been positioned for a general kind of exception (e.g.
 * the pre-defined \ref ccl_exception_exception "exception" kind). In this case
 * it can be useful to check what kind of exception has been raised. This test
 * is realized by \ref ccl_exception_is_raised.
 */
#ifndef __CCL_EXCEPTION_H__
# define __CCL_EXCEPTION_H__

# include <ccl/ccl-common.h>
# include <ccl/ccl-protos.h>

BEGIN_C_DECLS

# include <setjmp.h>

/*!
 * \brief Type of exception objects.
 */
typedef struct ccl_exception_st ccl_exception;

/*!
 * \brief Type encoding the kind of an exception.
 */
typedef struct ccl_exception_type_st ccl_exception_type;
typedef struct ccl_exception_local_variable_st ccl_exception_local_variable;

struct ccl_exception_st {
  const ccl_exception_type *type;
  jmp_buf context;
  ccl_exception_local_variable *vars;
  ccl_exception *next;
};

struct ccl_exception_type_st {
  const char *name;
  const ccl_exception_type *super;
};

/*!
 * \brief Dummy macro-function allowing to indicate exceptions thrown by 
 * a function.
 * \param _args_ space separated list of thrown exceptions 
 */
# define CCL_THROW(_args_) 

/*!
 * \brief Macro-function used to declare an exception \a exc. 
 * \param exc the name of the exception kind
 * \param super indicates that \a exc is a kind of (or inherits from) \a super.
 *        (for information purpose only, not actually used by the macro).
 */
# define CCL_DECLARE_EXCEPTION(exc,super)		\
  extern const ccl_exception_type CCL_EXCEPTION_NAME (exc)

/*!
 * \brief Macro-function used to define an new exception kind \a exc inheriting
 * from \a super kind.
 *
 */
# define CCL_DEFINE_EXCEPTION(exc,super)	 \
  const ccl_exception_type CCL_EXCEPTION_NAME (exc) = { \
    # exc, &CCL_EXCEPTION_NAME (super)		 \
  }

/*!
 * \brief Macro-function that build the symbol for exceptions used by
 * macro-functions for the definition and declaration of new exceptions
 * (see \ref CCL_DEFINE_EXCEPTION and \ref CCL_DECLARE_EXCEPTION).
 */
# define CCL_EXCEPTION_NAME(exc) ccl_exception_ ## exc

CCL_DECLARE_EXCEPTION (___toplevel_throwable, NULL);

/*!
 * \brief Top-level exception kind for errors.
 */
CCL_DECLARE_EXCEPTION (error, ___toplevel_throwable);

/*!
 * \brief Top-level exception kind for exceptions.
 */
CCL_DECLARE_EXCEPTION (exception, ___toplevel_throwable);

/*!
 * \brief Special \ref ccl_exception_error kind raised when an abnormal 
 * behavior is detected by the program.
 */
CCL_DECLARE_EXCEPTION (internal_error, error);

/*!
 * \brief Special \ref ccl_exception_exception raised during normal operation
 * of the program.
 */
CCL_DECLARE_EXCEPTION (runtime_exception, exception);

/*!
 * \brief Special \ref ccl_exception_exception raised when a division by zero 
 * occurs.
 */
CCL_DECLARE_EXCEPTION (division_by_zero, runtime_exception);

/*!
 * \brief Message of the last raised exception.
 * If no exception has been raised the content of this variable is undefined.
 */
extern const char *ccl_exception_current_message;

/*!
 * \brief Stacks a new context permitting to catch exception of kind \a _extype_
 *
 * Exception with (or inheriting of) the kind \a _extype_ can be caught and 
 * treated using the \ref ccl_catch statement or ignored with \ref ccl_no_catch.
 */
# define ccl_try(_extype_)					\
  do {								\
    ccl_exception here;						\
    ccl_exception_push (&here, &CCL_EXCEPTION_NAME (_extype_));	\
    if (setjmp (here.context) == 0)				\
      { 


/*! 
 * \brief Begins the treatment part of exception.
 * This statement is used after the code put into a \ref ccl_try context.
 * It is followed by a code that handle raised exception.  
 */
# define ccl_catch			       \
    ccl_exception_cleanup_local_variables (0); \
    ccl_exception_pop ();		       \
  } else {				       \
    ccl_exception_cleanup_local_variables (1); \
    ccl_exception_pop ();

# define ccl_catch_rethrow ccl_catch { ccl_rethrow (); } ccl_end_try

/*!
 * \brief Statement that MUST be used after the \ref ccl_catch statement to
 * correctly close a \ref ccl_try statement.
 */
# define ccl_end_try \
    }	\
  } while (0)

/*!
 * \brief Simple statement that simply catches (without other treatment) \a 
 * _extype_ exceptions thrown by the statement \a _stmt_.
 */
# define ccl_trycatch(_extype_,_stmt_)			\
  ccl_try (_extype_) _stmt_  ccl_catch ccl_end_try

/*!
 * \brief Statement used to ignore raised exceptions (up to the kind specified
 * in the \ref ccl_try statement). 
 * After this statement \ref ccl_end_try should not be used.
 */
# define ccl_no_catch	\
  }			\
  ccl_exception_pop (); \
} while (0)

/*!
 * \brief Throws an exception of type \a _extype.
 * The message \a _msg is associated with the exception and could be retrieve
 * via the \ref ccl_exception_current_message variable.
 *
 * If the exception is not caught by a \a ccl_try statement handling \a _extype
 * kind of exceptions, a message indicates the location of this throw and 
 * the program is aborted.
 */
# define ccl_throw(_extype,_msg) \
  ccl_throw__ (&CCL_EXCEPTION_NAME (_extype), _msg, __FILE__, __LINE__)

/*!
 * \brief Behaves like ccl_throw but the message is replaced by the name
 * of the exception.
 */
# define ccl_throw_no_msg(_extype) ccl_throw (_extype, # _extype)


/*!
 * \brief Throws an exception with same type than the last exception raised.
 * This function is useful in \a ccl_catch statements after a local treatment
 * of an exception.
 */
# define ccl_rethrow() ccl_rethrow__ (__FILE__, __LINE__)

/*!
 * \brief Checks if the last raised exception is a \a _e_ or inherits from this
 * type.
 */
# define ccl_exception_is_raised(_e_) \
  ccl_exception_is_raised__ (&CCL_EXCEPTION_NAME (_e_))

/*!
 * \brief Print a message on CCL_LOG_ERROR that locates the last exception
 * and its associated error message.
 */
extern void
ccl_exception_print (void);

/*
 * The following functions are used by the implementation 
 */

extern void
ccl_throw__ (const ccl_exception_type *ex, const char *msg,
	     const char *file, int line);

extern void
ccl_rethrow__ (const char *file, int line);

extern void
ccl_exception_push (ccl_exception *e, const ccl_exception_type *type);

extern void
ccl_exception_pop (void);


extern int
ccl_exception_is_raised__ (const ccl_exception_type *type);

# define ccl_exception_local_variable(del, pvar) \
  ccl_exception_local_variable__ ((ccl_delete_proc *)(del), (pvar))

extern void
ccl_exception_local_variable__ (ccl_delete_proc *del, void *pvar);

extern void
ccl_exception_cleanup_local_variables (int usedel);

END_C_DECLS

#endif /* ! __CCL_EXCEPTION_H__ */
