/*
 * ccl-heap.c -- a generic binary heap
 * 
 * This file is a part of the C Common Library (CCL) project. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the AltaRica Public License that comes with this
 * package.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 */

/*!
 * \file
 * \brief
 * 
 */
#include "ccl-assert.h"
#include "ccl-memory.h"
#include "ccl-array.h"
#include "ccl-heap.h"

struct ccl_heap_st
{  
  ccl_compare_with_data_func *cmp;
  void *data;
  int nb_elements;
  int init_size;
  CCL_ARRAY (void *) heap;
};

			/* --------------- */

static int
s_default_cmp (const void *p1, const void *p2);

static int
s_default_data_cmp (const void *p1, const void *p2, void *data);

static int
s_cmp_no_data (const void *p1, const void *p2, void *data);

#define CMP(h,e1,e2) ((h)->cmp((e1),(e2),(h)->data))

			/* --------------- */

ccl_heap *
ccl_heap_create (int size, ccl_compare_func *cmp)
{
  if (cmp == NULL)
    cmp = s_default_cmp;
  return ccl_heap_create_with_data (size, s_cmp_no_data, cmp);
}

			/* --------------- */

ccl_heap *
ccl_heap_create_with_data (int size, ccl_compare_with_data_func *cmp,
			   void *data)
{
  ccl_heap *result = ccl_new (ccl_heap);

  ccl_pre (size > 0);

  result->cmp = (cmp == NULL) ? s_default_data_cmp : cmp;
  result->data = data;
  result->nb_elements = 0;
  result->init_size = size;
  ccl_array_init_with_size (result->heap, size);

  return result;
}

void
ccl_heap_delete (ccl_heap *h)
{
  ccl_heap_clear_and_delete (h, NULL);
}

			/* --------------- */

void
ccl_heap_clear (ccl_heap *h, ccl_delete_proc *del)
{
  ccl_pre (h != NULL);

  if (del != NULL)
    {
      int i;

      for (i = 0; i < h->heap.size; i++)
	del (h->heap.data[i]);
    }

  ccl_array_trim (h->heap, h->init_size);
}


			/* --------------- */

void
ccl_heap_clear_and_delete (ccl_heap *h, ccl_delete_proc *del)
{
  ccl_heap_clear (h, del);
  ccl_array_delete (h->heap);
  ccl_delete (h);
}

			/* --------------- */

int
ccl_heap_get_size (const ccl_heap *h)
{
  ccl_pre (h != NULL);

  return h->nb_elements;
}

			/* --------------- */

void *
ccl_heap_get_first (const ccl_heap *h)
{
  ccl_pre (! ccl_heap_is_empty (h));

  return h->heap.data[0];
}

			/* --------------- */

#define heap_element(h, i) ((h)->heap.data[(i)-1])
#define left_child(i) ((i) << 1)
#define right_child(i) (((i) << 1) + 1)
#define parent(i) ((i) >> 1)


			/* --------------- */

#if 0
static void
s_check_structure (ccl_heap *h)
{

  int i;

  for (i = 1; i <= h->nb_elements; i++)
    {
      int l = left_child (i);
      int r = right_child (i);

      if (l <= h->nb_elements)
	ccl_assert (CMP (h, heap_element (h, i), heap_element (h,l)) <= 0);
      if (r <= h->nb_elements)
	ccl_assert (CMP (h, heap_element (h, i), heap_element (h,r)) <= 0);
    }
}
#else
#define s_check_structure(h) ((void) 0)
#endif
			/* --------------- */

static void
s_heap_build (ccl_heap *h, int i)
{
  int l, r, max, stop = 0;
  void *obj;
  
  while (! stop)
    {
      l = left_child (i);
      r = right_child (i);
      if (l <= h->nb_elements && 
	  CMP (h, heap_element (h, l), heap_element (h, i)) < 0)
	max = l;
      else
	max = i;
      if (r <= h->nb_elements && 
	  CMP (h, heap_element (h, r), heap_element (h, max)) < 0)
	max = r;
      stop = (max == i);
      if (! stop)
	{
	  obj = heap_element (h, i);
	  heap_element (h, i) = heap_element (h, max);
	  heap_element (h, max) = obj;
	  i = max;
	}
    }
  s_check_structure (h);
}

void *
ccl_heap_take_first (ccl_heap *h)
{
  void *result = ccl_heap_get_first (h);
  
  heap_element (h, 1) = heap_element (h, h->nb_elements);
  h->nb_elements--;
  s_heap_build (h, 1);

  return result;
}

			/* --------------- */

void
ccl_heap_add (ccl_heap *h, void *obj)
{
  int i;

  ccl_pre (h != NULL);  

  h->nb_elements++;
  ccl_array_ensure_size (h->heap, h->nb_elements); 
  i = h->nb_elements;
  while (i > 1 && CMP (h, heap_element (h, parent (i)), obj) > 0)
    {
      heap_element (h, i) = heap_element (h, parent (i));
      i = parent (i);
    }
  heap_element (h, i) = obj;
  s_check_structure (h);
}

			/* --------------- */

static int
s_has_rec (const ccl_heap *h, int i, void *obj)
{
  int cmp;
  
  if (i > h->nb_elements)
    return 0;
  cmp = CMP (h, heap_element (h, i), obj);
  if (cmp > 0)
    return 0;
  if (cmp == 0)
    return 1;

  return (s_has_rec (h, left_child (i), obj) ||
	  s_has_rec (h, right_child (i), obj));
}

int
ccl_heap_has (const ccl_heap *h, void *obj)
{  
  ccl_pre (h != NULL);  

  return s_has_rec (h, 1, obj);
}

			/* --------------- */

void
ccl_heap_copy (ccl_heap *dst, const ccl_heap *src)
{
  dst->cmp = src->cmp;
  dst->data = src->data;
  dst->nb_elements = src->nb_elements;
  dst->init_size = src->init_size;
  ccl_array_ensure_size (dst->heap, src->heap.size);
  ccl_memcpy (dst->heap.data, src->heap.data, src->heap.size * sizeof (void ));
}

void
ccl_heap_log (ccl_log_type log, ccl_heap *h, 
	      void (*logproc) (ccl_log_type log, void *obj, void *data),
	      void *data)
{
  int i;

  ccl_log (log, "[");
  for (i = 0; i < h->nb_elements; i++)
    {
      logproc (log, h->heap.data[i], data);
      if (i < h->nb_elements - 1)
	ccl_log (log, " ");
    }
  ccl_log (log, "]\n");
}

			/* --------------- */

static int
s_default_cmp (const void *p1, const void *p2)
{
  return ((intptr_t) p1) - ((intptr_t) p2);
}

static int
s_default_data_cmp (const void *p1, const void *p2, void *data)
{
  return s_default_cmp (p1, p2);
}

static int
s_cmp_no_data (const void *p1, const void *p2, void *data)
{
  return ((ccl_compare_func *)data) (p1, p2);
}
