/*
 * ccl-memory.c -- Memory allocators
 * 
 * This file is a part of the C Common Library (CCL) project. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the AltaRica Public License that comes with this
 * package.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 */

/*!
 * \file
 * \brief
 * 
 */
#include "ccl-assert.h"
#include <stdio.h>
#include "ccl-string.h"
#include "ccl-memory.h"

struct exhaustion_handler
{
  struct exhaustion_handler *next;
  ccl_memory_exhaustion_handler *proc;
  void *data;
};

#define MAX_NB_MODULES 1000
struct module
{
  const char *filename;
  size_t allocated;
  size_t nb_allocated;
  size_t released;
  size_t nb_released;
  size_t blockmaxsize;
  size_t maxsize;
  size_t realmem;
  int balanced;
};

struct memblock
{
  size_t size;
  struct module *mod;
  int line;
};

static int is_init = 0;
static struct module MODULES[MAX_NB_MODULES];
static struct module *lastmodule = &MODULES[MAX_NB_MODULES];


			/* --------------- */

#define TRY_ALLOCATION(_alloc)					\
  do								\
    {								\
      void *result = _alloc;					\
								\
      if (result == NULL)					\
	{							\
	  if (HANDLERS != NULL)					\
	    {							\
	      s_call_exhaustion_handlers ();			\
	      result = _alloc;					\
	      if (result != NULL)				\
		goto end;					\
	    }							\
	  ccl_throw_no_msg (memory_exhausted_exception);	\
	}							\
    end:							\
      return result;						\
    }								\
  while (0)

CCL_DEFINE_EXCEPTION (memory_exhausted_exception, runtime_exception);

			/* --------------- */

static struct exhaustion_handler *HANDLERS = NULL;

			/* --------------- */


static void
s_call_exhaustion_handlers (void)
{
  struct exhaustion_handler *H;

  for (H = HANDLERS; H; H = H->next)
    H->proc (H->data);
}

			/* --------------- */

void
ccl_memory_add_exhaustion_handler (ccl_memory_exhaustion_handler *hdl, 
				   void *data)
{
  struct exhaustion_handler *HDL = ccl_new (struct exhaustion_handler);

  HDL->next = HANDLERS;
  HDL->proc = hdl;
  HDL->data = data;
  HANDLERS = HDL;
}

			/* --------------- */

void
ccl_memory_del_exhaustion_handler (ccl_memory_exhaustion_handler *hdl, 
				   void *data)
{
  struct exhaustion_handler **pH;

  for (pH = &HANDLERS; *pH; pH = &((*pH)->next))
    {
      if ((*pH)->proc == hdl && (*pH)->data == data)
	{
	  struct exhaustion_handler *H = *pH;
	  *pH = H->next;
	  ccl_delete (H);
	  break;
	}
    }
}

			/* --------------- */

void *
ccl_real_malloc (size_t size)
{
  TRY_ALLOCATION (malloc (size));
}

			/* --------------- */

void
ccl_real_free (void *ptr)
{
  ccl_pre (ptr != NULL);

  free (ptr);
}

			/* --------------- */

void *
ccl_real_realloc (void *ptr, size_t size)
{
  TRY_ALLOCATION (realloc (ptr, size));
}

			/* --------------- */

void *
ccl_real_calloc (size_t nb_el, size_t el_size)
{
  TRY_ALLOCATION (calloc (nb_el, el_size));
}

			/* --------------- */

static struct module *
s_find_module (const char *filename)
{
  struct module *m;

  if (is_init == 0)
    {
      for (m = MODULES; m != lastmodule; m++)
	ccl_memzero (m, sizeof (*m));
      is_init = 1;
    }

  for (m = MODULES; m != lastmodule; m++)
    {
      if (m->filename == NULL)
	{	  
	  m->filename = filename;
	  m->balanced = 1;
	  return m;
	}
      else if (strcmp (m->filename, filename) == 0)
	return m;
    }
  fprintf (stderr, "ccl-memory: not enough modules have been defined.\n");
  abort ();
}


void *
ccl_debug_malloc (size_t size, const char *filename, int line)
{
  struct module *m = s_find_module (filename);
  size_t asize = size + sizeof (struct memblock);
  struct memblock *res = malloc (asize);

  res->size = size;
  res->mod = m;
  res->line = line;
  
  if (size > m->blockmaxsize)
    m->blockmaxsize = size;
  m->allocated += size;
  m->nb_allocated++;
  m->realmem += size;
  if (m->realmem > m->maxsize)
    m->maxsize = m->realmem;
  
  return res + 1;
}

void
ccl_debug_free (void *ptr, const char *filename, int line)
{
  struct module *m = s_find_module (filename);
  struct memblock *mblock = ptr;

  mblock--;

  if (mblock->mod != m)
    {
      fprintf (stderr, "warning: block at %p allocated in %s:%d released "
	       "in %s:%d\n",
	       ptr, mblock->mod->filename, mblock->line, filename, line);
      mblock->mod->balanced = 0;
      m->balanced = 0;
    }
  ccl_assert (ccl_imply (m->balanced, mblock->size <= m->blockmaxsize));
  ccl_assert (ccl_imply (m->balanced, mblock->size <= m->realmem));

  m->released += mblock->size;
  m->nb_released++;
  m->realmem -= mblock->size;

  ccl_real_free (mblock);
}

void *
ccl_debug_calloc (size_t nb_el, size_t size, const char *filename, int line)
{
  size_t asize = nb_el * size;
  void *res = ccl_debug_malloc (asize, filename, line);
  ccl_memzero (res, asize);
  
  return res;
}

void *
ccl_debug_realloc (void *p, size_t newsize, const char *filename, int line)
{
  void *res = ccl_debug_malloc (newsize, filename, line);
  
  if (p != NULL)
    {
      struct memblock *mblock = p;
      size_t blocksize = mblock[-1].size;
      size_t minsz = newsize < blocksize ? newsize : blocksize;
      ccl_memcpy (res, p, minsz);
      ccl_debug_free (p, filename, line);
    }
  
  return res;
}

#define PREFIX_FMT "%-20s= "

static void
s_display_memory_size (const char *prefix, size_t sz, const char *suffix)
{
  fprintf (stderr, PREFIX_FMT, prefix);
  if (sz < (1L <<10))
    fprintf (stderr, "%zu o", sz);
  else if (sz < (1L << 20))
    fprintf (stderr, "%4.2f Ko", (double) sz / (double) (1L << 10));
  else if (sz < (1L << 30))
    fprintf (stderr, "%4.2f Mo", (double) sz / (double) (1L << 20));
  else if (sizeof (size_t) == 8)
    fprintf (stderr, "%4.2f Go", (double) sz / (double) (1L << 30));
  else
    fprintf (stderr, "%4.2f Mo", (double) sz / (double) (1L << 20));
  fprintf (stderr, "%s", suffix);
}
		       
void
ccl_debug_log_statistics_per_module (void)
{
  struct module *m;
  size_t realmem = 0;
  size_t allocated = 0;
  size_t released = 0;
  size_t nb_active_blocks = 0;
    
  for (m = MODULES; m != lastmodule && m->filename != NULL; m++)
    {
      allocated += m->allocated;
      released += m->released;
      realmem += m->realmem;
      nb_active_blocks += m->nb_allocated - m->nb_released;
    }
  
  ccl_assert (allocated >= released);
  ccl_assert (realmem == allocated - released);
  
  for (m = MODULES; m != lastmodule && m->filename != NULL; m++)
    {
      double part = 100.0*(double) m->realmem /(double) realmem;
      
      fprintf (stderr, "%s (%sbalanced):\n",
	       m->filename, m->balanced ? "" : "un");
      s_display_memory_size ("  blockmaxsize", m->blockmaxsize, "\n");
      s_display_memory_size ("  maxsize",m->maxsize, "\n");
      s_display_memory_size ("  allocated", m->allocated, " #blocks = ");
      fprintf (stderr, "%zu\n", m->nb_allocated);      
      s_display_memory_size ("  released", m->released, " #blocks = ");
      fprintf (stderr, "%zu\n", m->nb_released);
      s_display_memory_size ("  realmem", m->realmem, "\n");
      fprintf (stderr, PREFIX_FMT, "  part/total");
      if (part < 0.01)
	fprintf (stderr, "< 0.01%%\n");
      else
	fprintf (stderr, "%3.2f %%\n", part);
    }
  s_display_memory_size ("addition memory",
			 sizeof (struct memblock) * nb_active_blocks,
			 " #blocks = ");
  fprintf (stderr, "%zu\n", nb_active_blocks);
  s_display_memory_size ("total memory in use", realmem, "\n");
  s_display_memory_size ("actual memory in use",
			 realmem + sizeof (struct memblock) * nb_active_blocks,
			 "\n");
}
