/*
 * ccl-memory.h -- Memory allocators
 * 
 * This file is a part of the C Common Library (CCL) project. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the AltaRica Public License that comes with this
 * package.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 */

/*!
 * \file ccl/ccl-memory.h
 * \brief Memory Allocators
 * 
 */
#ifndef __CCL_MEMORY_H__
# define __CCL_MEMORY_H__

# ifdef HAVE_CONFIG_H
#  include <config.h>
# endif
# include <ccl/ccl-common.h>
# include <ccl/ccl-exception.h>

BEGIN_C_DECLS

typedef void ccl_memory_exhaustion_handler (void *data);

/*!
 * \brief Exception raised on free memory exhaustion.
 */
CCL_DECLARE_EXCEPTION (memory_exhausted_exception, runtime_exception);

extern void
ccl_memory_add_exhaustion_handler (ccl_memory_exhaustion_handler *hdl,
				   void *data);

extern void
ccl_memory_del_exhaustion_handler (ccl_memory_exhaustion_handler *hdl,
				   void *data);

/*!
 * \brief Allocates a new uninitialized bloc of \a size bytes.
 * \param size the number of bytes to allocate
 */
# if CCL_DEBUG_MEMORY
#  define ccl_malloc(sz) ccl_debug_malloc(sz,__FILE__,__LINE__)
#  define ccl_calloc(nb,sz) ccl_debug_calloc(nb,sz,__FILE__,__LINE__)
#  define ccl_realloc(p,sz) ccl_debug_realloc(p,sz,__FILE__,__LINE__)
#  define ccl_free(p) ccl_debug_free(p,__FILE__,__LINE__)
test
# else
#  define ccl_malloc(sz) ccl_real_malloc(sz)
#  define ccl_calloc(nb,sz) ccl_real_calloc(nb,sz)
#  define ccl_realloc(p,sz) ccl_real_realloc(p,sz)
#  define ccl_free(p) ccl_real_free(p)
# endif

# define CCL_DECLARE_DELETE_PROC(_id) extern void _id (void *p)
# define CCL_DEFINE_DELETE_PROC(_id) void _id (void *p) { ccl_delete (p); }
  
/*!
 * \brief Allocates an array of \a _sz_ elements of type \a _type_.
 */
# define ccl_new_array(_type_,_sz_) \
  ((_type_ *)ccl_calloc (_sz_, sizeof(_type_)))

/*!
 * \brief Allocates a new block large enough to receive an object of type 
 * \a _type_
 */
# define ccl_new(_type_) ccl_new_array (_type_, 1)

/*!
 * \brief Releases the memory pointed by \a ptr.
 */
# define ccl_delete(_ptr_) ccl_free (_ptr_)

/*!
 * \brief Releases the memory pointed by \a ptr if not NULL.
 */
# define ccl_zdelete(_del,_ptr_)		\
  do						\
    {						\
      if ((_ptr_) != NULL)			\
	_del (_ptr_);				\
    }						\
  while(0)

/*!
 * \brief Fill the memory block at address \a _p with \a _size 0-bytes.
 */
# define ccl_memzero(_p,_size) ccl_memset (_p, 0, _size)

/*!
 * \brief Fill the memory block at address \a _p with \a _size bytes equal to
 * \a _c.
 */
# define ccl_memset(_p,_c,_size) memset (_p, _c, _size)

/*!
 * \brief Copy \a _size bytes from the address \a _src to the address \a _dst.
 */
# define ccl_memcpy(_dst,_src,_size) memcpy (_dst, _src, _size)

/*!
 * \brief Lexicographically compates memory blocks at addresses \a _p1 and 
 * \a _p2 for at most \a _size bytes.
 */
# define ccl_memcmp(_p1,_p2,_size) memcmp (_p1, _p2, _size)

extern void *
ccl_real_malloc (size_t size);
extern void
ccl_real_free (void *ptr);
extern void *
ccl_real_realloc (void *ptr, size_t size);
extern void *
ccl_real_calloc (size_t nb_el, size_t el_size); 

extern void *
ccl_debug_malloc (size_t size, const char *filename, int line);
extern void
ccl_debug_free (void *ptr, const char *filename, int line);
extern void *
ccl_debug_realloc (void *ptr, size_t size, const char *filename, int line);
extern void *
ccl_debug_calloc (size_t nb_el, size_t el_size, const char *filename,
		  int line); 
extern void
ccl_debug_log_statistics_per_module (void);

END_C_DECLS

#endif /* ! __CCL_MEMORY_H__ */
