/*
 * GENEPI is the GENEric Presburger programming Interface.
 * 
 * This file is a part of the GENEric Presburger programming Interface. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

/*!
 * \file genepi/genepi-trace.h
 * \brief Trace commands used by DIJO 
 * 
 * DIJO is used to reinterpret traces generated by the GENEPI solvers. 
 * When the traces are activated (using genepi_solver_set_trace_file()), GENEPI
 * generates a stream of lines. The very first line is the name of the plugin
 * used to generate the trace. Then, following lines are lists of integers 
 * separated by white spaces. Each line is a command to be executed by DIJO
 * and the first integer is the identifier of a command, following integers 
 * are its arguments. Variables are identified with integers. 
 */

#ifndef __GENEPI_TRACE_H__
# define __GENEPI_TRACE_H__

/*!
 * \brief Identifier of DIJO commands.
 */
typedef enum {
  /*!
   * \brief <b>syntax:</b> cmd id 
   *
   * Apply #genepi_set_add_reference to the variable \a id. The variable \a id
   * must exist.
   */
  GENEPI_TRACE_ADD_REF = 0,  

  /*!
   * \brief <b>syntax:</b> cmd id 
   *
   * Apply #genepi_set_del_reference to the variable \a id. The variable \a id
   * must exist.
   */
  GENEPI_TRACE_DEL_REF,

  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_top (width).
   * If the variable does not exist then it is created.
   */
  GENEPI_TRACE_TOP,      
    
  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_bot (width).
   * If the variable does not exist then it is created.
   */
  GENEPI_TRACE_BOT,

  /*!
   * \brief <b>syntax:</b> cmd dest_id cmp cst alpha_size alpha_1 ...
   *
   * Assign to the variable \a dest_id the result of 
   * #genepi_set_linear_operation ([alpha_1,...,alpha_{alpha_size}], cmp, 
   * cst). If the variable does not exist then it is created.
   */
  GENEPI_TRACE_LINEAR,

  /*!
   * \brief <b>syntax:</b> cmd dest_id idop1 idop2 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_union (idop1,
   * idop2). Variables \a idop1 and \a idop2 must exist.
   */
  GENEPI_TRACE_UNION,

  /*!\brief <b>syntax:</b> cmd dest_id idop1 idop2 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_intersection
   * (idop1, idop2). Variables \a idop1 and \a idop2 must exist.
   */
  GENEPI_TRACE_INTERSECTION,

   /*!
    * \brief <b>syntax:</b> cmd dest_id idop 
    *
    * Assign to the variable \a dest_id the result of #genepi_set_complement
    * (idop1, idop2). Variables \a idop1 and \a idop2 must exist.
    */
  GENEPI_TRACE_COMPLEMENT,

  /*!
   * \brief <b>syntax:</b> cmd dest_id idop sel_size sel_1 ...
   *
   * Assign to the variable \a dest_id the result of #genepi_set_project
   * (idop, [sel_1, ..., sel_{sel_size}]). The variable \a idop must exist.
   */
  GENEPI_TRACE_PROJECT,

  /*!
   * \brief <b>syntax:</b> cmd dest_id idop sel_size sel_1 ...
   *
   * Assign to the variable \a dest_id the result of #genepi_set_invproject
   * (idop, [sel_1, ..., sel_{sel_size}]). The variable \a idop must exist.
   */
  GENEPI_TRACE_INVPROJECT,

  /*!
   * \brief <b>syntax:</b> cmd dest_id idop1 idop2 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_apply
   * (idop1, idop2). Variables \a idop1 and \a idop2 must exist.
   */
  GENEPI_TRACE_APPLY,

  /*!
   * \brief <b>syntax:</b> cmd dest_id idop1 idop2 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_applyinv
   * (idop1, idop2). Variables \a idop1 and \a idop2 must exist.
   */
  GENEPI_TRACE_APPLYINV,

  /*!
   * \brief <b>syntax:</b> cmd cmp idop1 idop2 result 
   *
   * Check that the result of #genepi_set_compare (idop1, cmp, idop2) is \a 
   * result. Variables \a idop1 and \a idop2 must exist. \a cmp is the integer
   * value of an enumeration #genepi_comparator.
   */
  GENEPI_TRACE_CHK_COMPARE,

  /*!
   * \brief <b>syntax:</b> cmd idop result 
   *
   * Check that the result of #genepi_set_is_empty (idop) is \a result.
   * The variable \a idop must exist.
   */
  GENEPI_TRACE_CHK_EMPTY,

  /*!
   * \brief <b>syntax:</b> cmd idop result 
   *
   * Check that the result of #genepi_set_is_full (idop) is \a result.
   * The variable \a idop must exist.
   */
  GENEPI_TRACE_CHK_FULL,

  /*!
   * \brief <b>syntax:</b> cmd idop result 
   * 
   * Check that the result of #genepi_set_is_finite (idop) is \a result.
   * The variable \a idop must exist.
   */
  GENEPI_TRACE_CHK_FINITE,

  /*!
   * \brief <b>syntax:</b> cmd idop result 
   * 
   * Check that the result of #genepi_set_get_width (idop) is \a result.
   * The variable \a idop must exist.
   */
  GENEPI_TRACE_CHK_WIDTH,

  /*!
   * \brief <b>syntax:</b> cmd idop result 
   * 
   * Check that #genepi_set_get_data_structure_size (idop) is \a result.
   * This test is enabled only if the solver used by DIJO is the same than
   * the one used by GENEPI.
   * The variable \a idop must exist.
   */
  GENEPI_TRACE_CHK_DS_SIZE,

  /*!
   * \brief <b>syntax:</b> cmd result 
   * 
   * Check that the current number of created variable is \a result.
   */
  GENEPI_TRACE_CHK_NB_VARS,

  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_top_N (width).
   * If the variable \a dest_id does not exist then it is created.
   */
  GENEPI_TRACE_TOP_N,

  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_top_Z (width).
   * If the variable \a dest_id does not exist then it is created.
   */
  GENEPI_TRACE_TOP_Z,

  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_top_P (width).
   * If the variable \a dest_id does not exist then it is created.
   */
  GENEPI_TRACE_TOP_P,

  /*!
   * \brief <b>syntax:</b> cmd dest_id width 
   *
   * Assign to the variable \a dest_id the result of #genepi_set_top_R (width).
   * If the variable \a dest_id does not exist then it is created.
   */
  GENEPI_TRACE_TOP_R,

  /*!
   * \brief <b>syntax:</b> cmd idop result size xden x_1 ... 
   * 
   * Check the  \a result of #genepi_set_is_solution on \a idop and [x_1/xden,
   * ..., x_{size}/xden]. The variable idop must exist.
   */
  GENEPI_TRACE_CHK_SOLUTION
} genepi_trace_command;

#endif /* ! __GENEPI_TRACE_H__ */
