/*
 * batof-patterns.c -- add a comment about this file
 * 
 * This file is a part of the PresTAF project. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

/*!
 * \file
 * \brief
 * 
 */
#include <math.h>
#include <time.h>
#include <ccl/ccl-memory.h>
#include <ccl/ccl-stack.h>
#include <ccl/ccl-hash.h>
#include "prestaf-batof-p.h"

# define NO_DIM NULL
# define NO_U NULL
# define NO_V NULL
# define NO_X NULL
# define NO_Y NULL

typedef struct path_cell_st path_cell;
struct path_cell_st
{
  int b;
  path_cell *next;
};

typedef struct pattern_group_st pattern_group; 
struct pattern_group_st
{
  tla_matrix_t A;      /* matrix of size m x m */
  tla_group_t G; /* the group generated by the k vectors of size m stored in
		  * the columns of A 
		  */
  tla_matrix_t V;      /* matrix of size k x k such that A.V = HNF(A) */
  tla_vec_t *Zi; 
};
			/* --------------- */

static int
s_check_parity_and_product_of_ci_s (int dim, int *C, int *pCprod, int max, 
				    int *Cref);

static int *
s_attach_groups (sataf_ea *ea, int q0, int m, int **preds);

			/* --------------- */

static void
s_pattern_group_init (pattern_group *pg, int size, int dim);

static void
s_pattern_group_clear (pattern_group *pg);

static void
s_tla_group_to_pattern_group (pattern_group *R, tla_group_t G, int *C, 
			      int *Cref, tla_matrix_t *pX);

static void
s_pattern_group_successor (pattern_group *R, pattern_group *pg, int *C);

static int
s_pattern_group_compute_Zi (pattern_group *W0, pattern_group *W1, int *C);

static pattern_group *
s_compute_groups (tla_group_t G, int nb_states, int *C, int *pCprod, 
		  tla_matrix_t *pX);

			/* --------------- */

static int **
s_compute_preds (sataf_ea *ea);

static tla_vec_t *
s_compute_Aq (sataf_ea *ea, int **pred, int m);

static tla_vec_t **
s_compute_Zqb (sataf_ea *ea, int m, int k, pattern_group *groups,
	       int *agroups, int *C, int **preds);

static int **
s_compute_association_table (sataf_ea *ea, int m, int k, pattern_group *groups,
			     int *agroups, int *C, int Cprod, tla_vec_t **Zqb);

static int **
s_compute_patterns (sataf_ea *ea, int q0, int m, int k, int **atable, int *C, 
		    int Cprod, int *p_nb_patterns);

			/* --------------- */

int
btf_compute_patterns(tla_group_t G, sataf_ea *ea, int q0, int *C, 
		     int ***p_patterns, int *p_patwidth, int *p_nb_patterns,
		     tla_matrix_t *pX)
{
  int i;
  int Cprod = 1;
  int m = tla_group_get_size (G);
  int k = tla_group_get_dim (G);
  int nb_states = sataf_ea_get_nb_states (ea);  
  pattern_group *groups;
  int nb_groups;
  int is_presburger_set;

  CCL_DEBUG_START_TIMED_BLOCK (("compute patterns"));

  groups = s_compute_groups (G, nb_states, C, &Cprod, pX);
  nb_groups = tla_group_get_size (G);
  is_presburger_set = (groups != NULL);

  if (groups != NULL && ccl_debug_is_on)
    {
      int i;

      ccl_debug ("C=[ ");
      for (i = 0; i < k; i++)
	ccl_debug ("%d ", C[i]);
      ccl_debug ("]\nG=\n", i);
      tla_display_group (CCL_LOG_DEBUG, G);
      for (i = 0; i < m; i++)
	{
	  int j;
	  ccl_debug ("G%d=\n", i);
	  tla_display_group (CCL_LOG_DEBUG, groups[i].G);
	  ccl_debug ("A%d=\n", i);
	  tla_display_matrix (CCL_LOG_DEBUG, groups[i].A);
	  for (j = 0; j < k; j++)
	    {
	      ccl_debug ("Z%d,%d=", i, j);
	      tla_display_vec (CCL_LOG_DEBUG, groups[i].Zi[j]);
	      ccl_debug ("\n");
	    }
	}
    }

  if (is_presburger_set)
    { 
      int **atable;
      int **preds = s_compute_preds (ea);
      int *agroups = s_attach_groups (ea, q0, m, preds);
      tla_vec_t **Zqb = s_compute_Zqb (ea, m, k, groups, agroups, C, preds);

      is_presburger_set = (Zqb != NULL);
      if (is_presburger_set)
	{
	  atable = 
	    s_compute_association_table (ea, m, k, groups, agroups, C, Cprod, 
					 Zqb);
	  *p_patterns = 
	    s_compute_patterns (ea, q0, m, k, atable, C, Cprod, p_nb_patterns);
	  *p_patwidth = Cprod;

	  for (i = 2*nb_states*m-1; i >= 0; i--)
	    {
	      ccl_zdelete (ccl_delete, atable[i]);
	      if (Zqb[i] != NULL)
		{
		  tla_vec_clear ((*Zqb[i]));
		  ccl_delete (Zqb[i]);
		}
	    }
	  ccl_delete (Zqb);
	  ccl_delete (atable);
	}
      ccl_delete (agroups);
      for (i = 0; i < 2*nb_states; i++)
	ccl_delete (preds[i]);
      ccl_delete (preds);
    }

  /* clean-up all datas */
  if (groups != NULL)
    {
      for (i = 0; i < nb_groups; i++)
	s_pattern_group_clear (groups+i);
      ccl_delete (groups);
    }

  CCL_DEBUG_END_TIMED_BLOCK ();

  return is_presburger_set;
}

			/* --------------- */

static int
s_check_parity_and_product_of_ci_s (int dim, int *C, int *pCprod, int max, 
				    int *Cref)
{
  int i;
  int result = 1;
  mpz_t Ci_prod;

  mpz_init_set_ui (Ci_prod, 1);

  for (i = 0; i < dim && result; i++)
    {
      result = ((C[i] % 2) != 0);
      if (Cref != NULL)
	result = result && (C[i] == Cref[i]);

      mpz_mul_si (Ci_prod, Ci_prod, C[i]);
    }

  ccl_assert (mpz_fits_sint_p (Ci_prod));
  *pCprod = mpz_get_si (Ci_prod);
  result = result && (mpz_cmp_ui (Ci_prod, max ) <= 0);  

  mpz_clear (Ci_prod);

  return result;
}

			/* --------------- */

static int *
s_attach_groups (sataf_ea *ea, int q0, int m, int **preds)
{
  int nb_states = sataf_ea_get_nb_states (ea);
  int *result = ccl_new_array (int, nb_states * m);
  ccl_stack *todo = ccl_stack_create ();

  {
    int i;

    for (i = 0; i < nb_states * m; i++)
      result[i] = -1;
  }

  result[m * q0] = 0;
  ccl_stack_push (todo, (void *) (intptr_t) q0);
  ccl_stack_push (todo, (void *) (intptr_t) 0);
  while (!ccl_stack_is_empty (todo))
    {
      int b;
      int i = (intptr_t) ccl_stack_pop (todo);
      int q = (intptr_t) ccl_stack_pop (todo);
      int g = result[m * q + i];
      int i_qp = (m + i - 1) % m;
      int g_qp = (g + 1) % m;

      ccl_assert (g >= 0);

      for (b = 0; b < 2; b++)
	{
	  int *q_preds = preds[2 * q + b];
	  int nb_preds = q_preds[0];
	  int p;

	  for (p = 0; p < nb_preds; p++)
	    {
	      int qp = q_preds[p+1];

	      if (result[m * qp + i_qp] < 0)
		{
		  result[m * qp + i_qp] = g_qp;
		  ccl_stack_push (todo, (void *) (intptr_t) qp);
		  ccl_stack_push (todo, (void *) (intptr_t) i_qp);
		}
	      else
		ccl_assert (result[m  * qp + i_qp] == g_qp);
	    }
	}
    }

  ccl_stack_delete (todo);

  return result;
}

			/* --------------- */

static void
s_pattern_group_init (pattern_group *pg, int size, int dim)
{
  int i;

  tla_group_init (pg->G, size);
  tla_matrix_init (pg->A, size, dim);
  tla_matrix_init (pg->V, dim, dim);

  pg->Zi = ccl_new_array (tla_vec_t, dim);
  for (i = 0; i < dim; i++)
    tla_vec_init (pg->Zi[i], size);
}

			/* --------------- */

static void
s_pattern_group_clear (pattern_group *pg)
{
  int i;
  int dim = tla_group_get_dim (pg->G);

  tla_matrix_clear (pg->A);
  tla_matrix_clear (pg->V);

  for (i = 0; i < dim; i++)
    tla_vec_clear (pg->Zi[i]);
  ccl_delete (pg->Zi);

  tla_group_clear (pg->G);  
}

/* 
 * 1) Computation of the SNF of the Hermite matrix of G. Store the X matrix.
 *    If c != NULL we put diagonal coefficients S[i,i] into c[i] for i = 0, 
 *    ..., G->dim-1.
 * 2) Put the G->dim first lines of X as column of R->A
 *    Add k-th column of A as a generator of R->G 
 * 3) Compute the matrix R->V such that R->A x V = HNF(A)
 */
static void
s_tla_group_to_pattern_group (pattern_group *R, tla_group_t G, int *C, 
			      int *Cref, tla_matrix_t *pX)
{
  int j;
  tla_vec_t v;
  const int size = tla_group_get_size (G);
  const int dim = tla_group_get_dim (G);
  int m = size;
  int n = m+1;
  int i = m*n+m*m;
  tla_matrix_t S, X;

  /* init step */
  tla_matrix_init (S, m, n);
  tla_matrix_init (X, m, m);

  tla_vec_init (v, size);

  /* step 1 */
  tla_matrix_compute_snf (G->Hermite, S, NO_DIM, X, NO_Y, NO_U, NO_V);

  if (ccl_debug_is_on)
    {
      ccl_debug ("G->Hermite=\n");
      tla_display_matrix (CCL_LOG_DEBUG, G->Hermite);
      ccl_debug ("\n");

      ccl_debug ("S=\n");
      tla_display_matrix (CCL_LOG_DEBUG, S);
      ccl_debug ("\n");
    }

  if (C != NULL)
    {
      for (i = 0; i < dim; i++)
	{
	  ccl_assert (mpz_fits_sint_p (MM (S, i + 1, i + 1)));
	  C[i] = mpz_get_si (MM (S, i + 1, i + 1));
	  while ((C[i] & 0x1) == 0)
	    C[i] >>= 1;
	}
    }  

    
  /* step 2 */
  tla_matrix_resize (S, m, m);

  if (pX != NULL)
    tla_matrix_set (*pX, X);

  for (j = 1; j <= dim; j++)
    {
      for (i = 1; i <= size; i++)
	{	
	  /*
	  mpz_mul_ui (M (R->A, m, dim, i, j), M (X, m, m, i, j), Cref[j - 1]);
	  mpz_mul_ui (M (S, m, m, i, j), M (X, m, m, i, j), Cref[j - 1]);
	  */	  
	  mpz_set (MM (R->A, i, j), MM (X, i, j));
	  mpz_set (MM (S, i, j), MM (X, i, j));	 
	}
      tla_matrix_get_column (R->A, j, v);
      tla_group_add_generator (R->G, R->G, v);
    }

  if (ccl_debug_is_on)
    {
      ccl_debug ("X=\n");
      tla_display_matrix (CCL_LOG_DEBUG, X);
      ccl_debug ("\n");
    }

  /* step 3 */
  tla_matrix_compute_hnf (S, S, NULL, NULL, X);

  if (ccl_debug_is_on)
    {
      ccl_debug ("S=\n");
      tla_display_matrix (CCL_LOG_DEBUG, S);
      ccl_debug ("\n");
    }
  
  for (i = 1; i <= dim; i++)
    {
      int j;
      
      for (j = 1; j <= dim; j++)
	mpz_set (MM (R->V, i, j), MM (X, i, j));
    }

  /* clean up step */
  tla_matrix_clear (S);
  tla_matrix_clear (X);
  tla_vec_clear (v);
}

/*
 * 1) compute the auxiliary group \a Gaux generated by
 *    \{ gamma_0(pg->A[0]), ..., gamma_0(pg->A[k-1]) \} where pg->A[i] is the 
 *    i-th column of pg->A and k the dimension of pg->G.
 * 2) compute the result R from Gaux.
 */
static void
s_pattern_group_successor (pattern_group *R, pattern_group *pg, int *C)
{
  int i;
  tla_vec_t v;
  tla_vec_t gamma_v;
  tla_group_t Gaux;
  int m = tla_group_get_size (pg->G);
  int k = tla_group_get_dim (pg->G);

  /* init step */
  tla_group_init (Gaux, m);
  tla_vec_init (v, m);
  tla_vec_init (gamma_v, m);

  /* step 1 */
  for (i = 0; i < k; i++)
    {
      tla_matrix_get_column (pg->A, i+1, v);
      tla_vec_gamma(gamma_v, v, 0);
      tla_vec_z_mul_si(v, C[i], gamma_v);
      tla_group_add_generator (Gaux, Gaux, v);
    }

  /* step 2 */
  s_tla_group_to_pattern_group (R, Gaux, NULL, C, NULL);

  /* clean up step */
  tla_vec_clear (v);
  tla_vec_clear (gamma_v);
  tla_group_clear (Gaux);
}

/*
 * Compute Z_i vectors of W0. W1 is the successor of W0.
 * 1) for each column of v_i = W0->A[i], compute the vectors of lambdas
 *    obtained from the call to tla_group_has (W1, gamma0(v_i), lambdas).
 *  1.1) if gamma0(v_i) doesn't belong to W1 then 0 is returned.
 *  1.2) else, apply W1->V to the vector of lambdas. in order to 
 *       obtain W0->Zi[i] we keep the remainder of each component j by c[j].
 */
static int
s_pattern_group_compute_Zi (pattern_group *W0, pattern_group *W1, int *C)
{
  int i;
  int is_presburger_set = 1;
  int m = tla_group_get_size (W0->G);
  int k = tla_group_get_dim (W1->G);
  tla_vec_t v_i;
  tla_vec_t gamma_v_i;
  tla_vec_t lambdas;

  /* init step */
  tla_vec_init (v_i, m);
  tla_vec_init (gamma_v_i, m);
  tla_vec_init (lambdas, m);
  
  /* step 1 */
  for (i = 0; i < k && is_presburger_set; i++)
    {
      tla_matrix_get_column (W0->A, i + 1, v_i);
      tla_vec_gamma(gamma_v_i, v_i, 0);
      is_presburger_set = tla_group_has (W1->G, gamma_v_i, lambdas);
      if (is_presburger_set)
	{
	  int j;
	  tla_vec_t *z_i = &(W0->Zi[i]);

	  /* step 1.2 */
	  tla_matrix_mul_vector (*z_i, W1->V, lambdas);
	  for (j = 0; j < k; j++)
	    mpz_mod_ui ((*z_i)->num[j], (*z_i)->num[j], abs (C[j]));
	}
    }

  /* clean up step */
  tla_vec_clear (v_i);
  tla_vec_clear (lambdas);
  tla_vec_clear (gamma_v_i);
    
  return is_presburger_set;
}

/*
 * 1) compute initial pattern group W_0 from G and associated C_i's
 * 2) Check properties of C_i's due to Presburger context:
 *    . each C_i is an odd number.
 *    . the product of C_i's is <= to the total number of states.
 * 3) for i from 1 to m-1 compute W_i from W_{i-1} and associated Zi vectors.
 *    for W_{m-1} compute Zi vectors from W_0.
 */
static pattern_group *
s_compute_groups (tla_group_t G, int nb_states, int *C, int *pCprod, 
		  tla_matrix_t *pX)
{
  int i;
  int m = tla_group_get_size (G);
  int k = tla_group_get_dim (G);
  int is_presburger_set = 1;
  pattern_group *result;

  /* init step */
  result = ccl_new_array (pattern_group, m);
  for (i = 0; i < m; i++)
    s_pattern_group_init (result+i, m, k);

  /* step 1 */
  s_tla_group_to_pattern_group (result, G, C, C, pX);
  /* step 2 */
  is_presburger_set = 
    s_check_parity_and_product_of_ci_s (k, C, pCprod, nb_states, NULL);
  /* step 3 */
  for (i = 1; i < m && is_presburger_set ; i++)
    {
      s_pattern_group_successor (result+i, result+(i-1), C);
      is_presburger_set = 
	s_pattern_group_compute_Zi (result+(i-1), result+i, C);
    }
  is_presburger_set = (is_presburger_set && 
		       s_pattern_group_compute_Zi (result+(m-1), result, C));

  /* clean up step */
  if (!is_presburger_set)
    {
      for (i = 0; i < m; i++)
	s_pattern_group_clear (result+i);
      ccl_delete (result);
      result = NULL;
    }

  return result;
}

/*
 * This function computes the predecessor relation. Each entry 2*i+b of the 
 * returned array is the set of predecessors of the state \a i whose reach i
 * by the letter b. Each set of predecessors is an array of \f$n+1\$ cells 
 * where \f$n\f$ is the number of predecessors; the number \f$n\f$ is stored in
 * the first cell of each predecessor set.
 */
static int **
s_compute_preds (sataf_ea *ea)
{
  int s;
  int nb_states = sataf_ea_get_nb_states (ea);
  int **result = ccl_new_array (int *, 2*nb_states);
  
  for (s = 0; s < 2*nb_states; s++)
    {
      result[s] = ccl_new (int);
      result[s][0] = 0;
    }

  for (s = 0; s < nb_states; s++)
    {
      int b;

      for (b = 0; b < 2; b++)
	{
	  int nb_preds;
	  int *preds;
	  int succ = sataf_ea_get_successor (ea, s, b);

	  if (sataf_ea_is_exit_state (succ))
	    continue;
	  succ = sataf_ea_decode_succ_state (succ);

	  preds = result[2 * succ + b];
	  nb_preds = preds[0];
	  preds = ccl_realloc (preds, sizeof (int) * (nb_preds + 2));
	  result[2 * succ + b] = preds;
	  preds[1 + nb_preds] = s;
	  nb_preds++;
	  preds[0] = nb_preds;
	}
    }

  return result;
}

/* 
 * Compute for each state of the EA a vector corresponding to an accepted word.
 * The function proceeds as follow:
 * . first, the vector zero is associated to each accepting state.
 * . then, using a backward algorithm, we associate to each predecessor 
 *   s' of the already computed state s the vector gamma(vec_s, 0).
 */
static tla_vec_t *
s_compute_Aq (sataf_ea *ea, int **preds, int m)
{
  int s;
  int nb_states = sataf_ea_get_nb_states (ea);
  tla_vec_t tmp;
  tla_vec_t *result = ccl_new_array (tla_vec_t, nb_states);
  ccl_bittable *done = ccl_bittable_create (nb_states);
  ccl_bittable *todo = ccl_bittable_create (nb_states);

  tla_vec_init (tmp, m);

  for (s = 0; s < nb_states; s++)
    {
      tla_vec_init (result[s], m);
      if (sataf_ea_is_final (ea, s))	
	{
	  ccl_bittable_set (done, s);
	  ccl_bittable_set (todo, s);
	}
    }

  while ((s = ccl_bittable_get_first (todo)) >= 0)
    {
      int b;

      ccl_bittable_unset (todo, s);
      for (b = 0; b < 2; b++)
	{
	  int p;
	  int *preds_s = preds[2 * s + b];
	  int nb_preds = preds_s[0];
	  tla_vec_gamma (tmp, result[s], b);

	  for (p = 0; p < nb_preds; p++)
	    {
	      int pred = preds_s[p + 1];
	      
	      if (ccl_bittable_has (done, pred))
		continue;
	      
	      ccl_bittable_set (done, pred);
	      ccl_bittable_set (todo, pred);
	      tla_vec_set (result[pred], tmp);
	    }      
	}
    }

  ccl_assert (ccl_bittable_get_nb_one (done) == nb_states); 

  if (0 && (ccl_debug_is_on))
    {
      for (s = 0; s < nb_states; s++)
	{
	  ccl_debug ("Aq[%d]=", s);
	  tla_display_vec (CCL_LOG_DEBUG, result[s]);
	  ccl_debug ("\n");
	}
    }

  ccl_bittable_delete (done);  
  ccl_bittable_delete (todo);  
  tla_vec_clear (tmp);

  return result;
}

/*
 * Compute Zqb vectors.
 * 1) compute for each state q of ea the vector Aq[q]
 * 2) For each accessible triplet (q,b,i) compute a vector Zqb as follow:
 *  2.1) compute v=gamma0 (Aq[q'])-Aq[q]
 *  2.2) check if v belong to the group Gqi associated to (q,i).
 *  2.3) compute Zqb[m*q+i] from v by applying Gqi->V to lambdas obtained from 
 *       2.2 and reduced to remainders by c[i].
 * The accessibility of a triplet is checked with the existence a positive or
 * null group index in agroups[m*q+i].
 */
static tla_vec_t **
s_compute_Zqb (sataf_ea *ea, int m, int k, pattern_group *groups, int *agroups,
	       int *C, int **preds)
{
  int i;
  int q;
  int is_presburger_set;
  tla_vec_t *Aq;
  int nb_states;
  tla_vec_t **Zqb;
  tla_vec_t lambdas;
  tla_vec_t gamma_aqp;

  /* init step */
  is_presburger_set = 1;
  nb_states = sataf_ea_get_nb_states (ea);
  Zqb = ccl_new_array (tla_vec_t *, m*nb_states*2);
  /*
  for (i = m*nb_states*2-1; i >= 0; i--)
    tla_vec_init (Zqb[i], m);
  */

  tla_vec_init(lambdas, m);
  tla_vec_init(gamma_aqp, m);

  /* step 1 : */
  Aq = s_compute_Aq (ea, preds, m);

  /* step 2 : */
  for (q = 0; q < nb_states && is_presburger_set; q++)
    {
      for (i = 0; i < m && is_presburger_set; i++)
	{
	  int b;
	  int g = agroups[m * q + i];
	  pattern_group *pg = groups+g;

	  if (g < 0)
	    continue;

	  for (b = 0; b < 2 && is_presburger_set; b++)
	    {
	      int qp = sataf_ea_get_successor (ea, q, b);

	      if (sataf_ea_is_exit_state (qp))
		continue;
	      qp = sataf_ea_decode_succ_state (qp);

	      /* step 2.1 */
	      tla_vec_gamma (gamma_aqp, Aq[qp], b);
	      tla_vec_sub(gamma_aqp, gamma_aqp, Aq[q]);

	      /* step 2.2 */
	      is_presburger_set = tla_group_has (pg->G, gamma_aqp, lambdas);
	      ccl_assert (is_presburger_set);

	      if (is_presburger_set)
		{
		  int j;
		  tla_vec_t *zqb = Zqb[2 * (m * q + i) + b] = 
		    ccl_new (tla_vec_t);
		  tla_vec_init (*zqb, m);

		  /* step 2.3 */
		  tla_matrix_mul_vector (*zqb, pg->V, lambdas);
		  for (j = 0; j < k; j++)
		    mpz_mod_ui ((*zqb)->num[j], (*zqb)->num[j], abs(C[j]));
		}
	    }
	}
    }
  
  /* clean up step */
  tla_vec_clear(lambdas);
  tla_vec_clear(gamma_aqp);

  for (i = 0; i < nb_states; i++)
    tla_vec_clear (Aq[i]);
  ccl_delete (Aq);

  if (! is_presburger_set)
    {
      for (i = m*nb_states*2-1; i >= 0; i--)
	{
	  if( Zqb[i] != NULL)
	    {
	      tla_vec_clear (*(Zqb[i]));
	      ccl_delete (Zqb[i]);
	    }
	}
      ccl_delete (Zqb);
      Zqb = NULL;
    }

  return Zqb;
}

			/* --------------- */

static void
s_generate_alpha_vector_rec (int i, int *table, int *ic, int *alphas, 
			     tla_vec_t *Zqb, pattern_group *W, tla_vec_t x)
{
  if (i == W->G->dim)
    {
      int j;
      int k = W->G->dim;
      int src_counter;
      int tgt_counter;

      tla_vec_set (x, *Zqb);

      for (j = 0; j < k; j++)
	{
	  tla_vec_add_si_mul (x, x, alphas[j], W->Zi[j]);
	}

      for (j = 0; j < k; j++)
	mpz_mod_ui(x->num[j], x->num[j], ic[j]); 

      src_counter = mpz_get_si (x->num[k-1]);
      tgt_counter = alphas[k-1];
      
      for (j = k-2; j >= 0; j--)
	{
	  src_counter = mpz_get_si (x->num[j]) + ic[j]*src_counter;
	  tgt_counter = alphas[j] + ic[j]*tgt_counter;
	}
      table[src_counter] = tgt_counter;
    }
  else
    {
      int alpha;

      for (alpha = 0; alpha < ic[i]; alpha++)
	{
	  alphas[i] = alpha;
	  s_generate_alpha_vector_rec (i+1, table, ic, alphas, Zqb, W, x);
	}
    }
}

			/* --------------- */

static int **
s_compute_association_table (sataf_ea *ea, int m, int k, pattern_group *groups,
			     int *agroups, int *C, int Cprod, tla_vec_t **Zqb)
{
  int q;
  int **result;
  int *alphas;
  int table_size = Cprod;
  int nb_states = sataf_ea_get_nb_states (ea);
  tla_vec_t aux;

  CCL_DEBUG_START_TIMED_BLOCK (("compute association table... "));

  tla_vec_init (aux, m);

  alphas = ccl_new_array (int, k);
  result = ccl_new_array (int *, nb_states * m * 2);

  for (q = 0; q < nb_states; q++)
    {
      int i;

      for (i = 0; i < m; i++)
	{
	  int b;
	  int g_q  = agroups[m * q + i];
	  int iqp = (i + 1) % m;

	  if (g_q < 0)
	    continue;

	  for (b = 0; b < 2; b++)
	    {
	      int g_qp;
	      pattern_group *pg_qp;
	      int index;
	      int qp = sataf_ea_get_successor (ea, q, b);
	      
	      if (sataf_ea_is_exit_state (qp))
		continue;
	      qp = sataf_ea_decode_succ_state (qp);
	      g_qp = agroups[m * qp + iqp];
	      pg_qp = groups + g_qp;
	      index = 2 * (m * q + i) + b;

	      result[index] = ccl_new_array (int, table_size);
	      {
		int k;
		
		for (k = 0; k < table_size; k++)
		  result[index][k] = -1;
	      }
	      s_generate_alpha_vector_rec (0, result[index], C, alphas, 
					   Zqb[index], pg_qp, aux);

	      {
		int z;
		for (z = 0; z < table_size; z++)
		  ccl_assert (result[index][z] >= 0);	 
	      }
	    }
	}
    }

  tla_vec_clear (aux);
  ccl_delete (alphas);
  CCL_DEBUG_END_TIMED_BLOCK ();

  return result;
}

			/* --------------- */


static int **
s_compute_patterns (sataf_ea *ea, int q0, int m, int k, int **atable, int *C, 
		    int Cprod, int *p_nb_patterns)
{
  int **patterns;

  CCL_DEBUG_START_TIMED_BLOCK (("compute patterns... "));

  {
    int x;
    int current = 0;
    int table_size = Cprod;
    int nb_states = sataf_ea_get_nb_states (ea);  
    ccl_list *list = ccl_list_create ();
    ccl_bittable *visited = ccl_bittable_create (nb_states * m * Cprod);
    patterns = ccl_new_array (int *, m * nb_states);

    for (x = 0; x < table_size; x++)
      {
	if (ccl_bittable_has (visited, Cprod * (m * q0) + x))
	  continue;

	ccl_bittable_set (visited, Cprod * (m * q0) + x);

	ccl_list_add (list, (void *) (intptr_t) q0);
	ccl_list_add (list, (void *) (intptr_t) 0);
	ccl_list_add (list, (void *) (intptr_t) x);

	while (! ccl_list_is_empty (list))
	  {
	    int b;
	    int *partition;
	    int q = (intptr_t) ccl_list_take_first (list);
	    int i = (intptr_t) ccl_list_take_first (list);
	    int p = (intptr_t) ccl_list_take_first (list);
	    int i_qp = (i + 1) % m;

	    if (patterns[m * q + i] == NULL)
	      {
		int k;
		patterns[m * q + i] = ccl_new_array (int, Cprod);

		for (k = 0; k < Cprod; k++)
		  patterns[m * q + i][k] = -1;
	      }

	    partition = patterns[m * q + i];
	    partition[p] = current;

	    for (b = 0; b < 2; b++)
	      {
		int *table = atable[2 * (m * q + i) + b];
		int qp = sataf_ea_get_successor (ea, q, b);
		int index;

		if (sataf_ea_is_exit_state (qp))
		  continue;

		qp = sataf_ea_decode_succ_state (qp);

		index = Cprod * (m * qp + i_qp) + table[p];
		if (ccl_bittable_has (visited, index))
		  continue;

		ccl_bittable_set (visited, index);
		ccl_list_add (list, (void *) (intptr_t) qp);
		ccl_list_add (list, (void *) (intptr_t) i_qp);
		ccl_list_add (list, (void *) (intptr_t) table[p]);
	      }
	  }
	current++;
      }
    *p_nb_patterns = current;

    ccl_list_delete (list);
    ccl_bittable_delete (visited);
  }

  CCL_DEBUG_END_TIMED_BLOCK ();

  return patterns;
}

			/* --------------- */

/* 
 * intpoint = p[0] + C[0] * (p[1] + C[1] * (... (p[k-2] + C[k-2] * p[k-1])))
 */
void
btf_decompose_int_point (int *p, int intpoint, int *C, int k)
{
  int i;
  
  for (i = 0; i < k-1; i++)
    {
      p[i] = intpoint % C[i];
      intpoint = (intpoint - p[i]) / C[i];
    }

  p[i] = intpoint;
}

